package org.cs.snc;


import javafx.embed.swing.SwingFXUtils;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.fxml.Initializable;
import javafx.scene.control.*;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.layout.AnchorPane;
import javafx.stage.FileChooser;
import javafx.stage.Stage;

import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.ResourceBundle;


public class ConfigSceneController implements Initializable {

    // @FXML tag is required for any JavaFX GUI element that is
    // initialized at FXML load time instead of in the Java controller class
    @FXML private AnchorPane anchorPane;
    @FXML private AnchorPane configAnchorPane;
    @FXML private AnchorPane crossoverAnchorPane;
    @FXML private Button btnImport;
    @FXML private ImageView imgViewImported;

    @FXML private RadioButton radioBtnTriangles;
    @FXML private RadioButton radioBtnQuads;

    @FXML private ToggleButton toggleOnElitism;
    @FXML private ToggleButton toggleOffElitism;

    @FXML private ToggleButton toggleOnInheritance;
    @FXML private ToggleButton toggleOffInheritance;

    @FXML private Slider sliderNumOfIndividuals;
    @FXML private Slider sliderNumOfPolygons;
    @FXML private Slider sliderMutationChance;
    @FXML private Slider sliderCrossoverPoint;

    @FXML private Label labelNumOfIndividuals;
    @FXML private Label labelNumOfPolygons;
    @FXML private Label labelMutationChance;
    @FXML private Label labelCrossoverPoint;

    private File selectedFile;
    private BufferedImage importedBuffIm;

    private int numOfIndividualsPerGeneration = 75;
    private int numOfPolygonsPerIndividual = 150;
    private int numOfVertices = 3;

    private double mutationChance = 0.05;
    private double crossoverRatio = 0.5;
    private boolean randomInheritance = false;
    private boolean elitism = true;


    /* ******************************************************************************************* */
    // Function Name: initialize                                                                   //
    // Return Type: void                                                                           //
    // Parameters: URL, ResourceBundle                                                             //
    // Description: The initialize function is required to override when implementing the JavaFX   //
    //              Initializable class. A Java class typically implements the Initializable class //
    //              if it is the controller of an fxml file. The initialize function is called     //
    //              once the contents of the fxml file that the Java class controls have been      //
    //              loaded. In this function, all user input GUI elements are disabled until the   //
    //              user imports a photo. Also, slider min/max values are set, button labels are   //
    //              set, and listeners are created.                                                //
    /* ******************************************************************************************* */
    @Override
    public void initialize(URL url, ResourceBundle resourceBundle) {
        configAnchorPane.setVisible(false);

        createSliderValues();
        createPolygonRadioBtns();
        createElitismButtons();
        createRandomInheritanceButtons();
    }


    /* ******************************************************************************************* */
    // Function Name: btnImportClicked                                                             //
    // Return Type: void                                                                           //
    // Parameters: javafx.event.ActionEvent                                                        //
    // Description: This function is linked to a button via ConfigSceneView.fxml. When clicked,    //
    //              a dialog box pops up that allows the user to pick an image from their hard     //
    //              drive/SSD. I the user actually selects an image instead of pressing cancel,    //
    //              the file opened is converted to a java.awt.image.BufferedImage and saved for   //
    //              later, and also converted to a javafx.scene.image.Image for displaying to the  //
    //              screen.                                                                        //
    /* ******************************************************************************************* */
    @FXML
    public void btnImportClicked(ActionEvent actionEvent) {
        Stage stage = (Stage)((Button)actionEvent.getSource()).getScene().getWindow();
        if(openValidFile(stage)) {
            if(convertToImage()) {
                btnImport.setText("Import New Image");
                configAnchorPane.setVisible(true);
            }
        }
    }

    /* ******************************************************************************************* */
    // Function Name: btnNextClicked                                                               //
    // Return Type: void                                                                           //
    // Parameters: javafx.event.ActionEvent                                                        //
    // Description: This function is linked to a button via ConfigSceneView.fxml. When clicked,    //
    //              the next scene is loaded to the screen with FXMLLoader.load. All of the        //
    //              parameters that are picked by the user are saved to an AlgorithmData variable. //
    /* ******************************************************************************************* */
    @FXML
    public void btnNextClicked(ActionEvent actionEvent) throws IOException {
        Stage window = (Stage)((Button)actionEvent.getSource()).getScene().getWindow();
        window.setResizable(false);

        FXMLLoader fxmlLoader = new FXMLLoader(App.class.getResource("fxml/RunSceneView.fxml"));
        AnchorPane pane = fxmlLoader.load();

        RunSceneController runSceneController = fxmlLoader.getController();
        AlgorithmData data = new AlgorithmData(numOfIndividualsPerGeneration, numOfPolygonsPerIndividual, numOfVertices, mutationChance, crossoverRatio, randomInheritance, elitism, importedBuffIm);
        runSceneController.transferData(data);

        anchorPane.getChildren().setAll(pane);
    }


    /* ******************************************************************************************* */
    // Function Name: createSliderValues                                                           //
    // Return type: void                                                                           //
    // Parameters: None                                                                            //
    // Description: This function is called by initialize to set all of the min and max values for //
    //              the sliders. The listeners for the sliders are also set.                       //
    /* ******************************************************************************************* */
    private void createSliderValues() {
        labelNumOfIndividuals.setText("" + numOfIndividualsPerGeneration);
        labelNumOfPolygons.setText("" + numOfPolygonsPerIndividual);
        labelMutationChance.setText(String.format("%.1f", (100 * mutationChance)) + "%");
        labelCrossoverPoint.setText(String.format("%.0f", (100 * crossoverRatio)) + "%");

        sliderNumOfIndividuals.setMin(25);
        sliderNumOfIndividuals.setMax(200);
        sliderNumOfIndividuals.setValue(numOfIndividualsPerGeneration);
        sliderNumOfIndividuals.valueProperty().addListener((observableValue, oldValue, newValue) -> {
            numOfIndividualsPerGeneration = newValue.intValue();
            labelNumOfIndividuals.setText("" + newValue.intValue());
        });

        sliderNumOfPolygons.setMin(50);
        sliderNumOfPolygons.setMax(400);
        sliderNumOfPolygons.setValue(numOfPolygonsPerIndividual);
        sliderNumOfPolygons.valueProperty().addListener((observableValue, oldValue, newValue) -> {
            numOfPolygonsPerIndividual = newValue.intValue();
            labelNumOfPolygons.setText("" + newValue.intValue());
        });

        sliderMutationChance.setMin(0.01);
        sliderMutationChance.setMax(0.1);
        sliderMutationChance.setValue(mutationChance);
        sliderMutationChance.valueProperty().addListener((observableValue, oldValue, newValue) -> {
            mutationChance = Math.round(newValue.doubleValue() * 1000.0) / 1000.0;
            labelMutationChance.setText(String.format("%.1f", (mutationChance * 100)) + "%");
        });

        sliderCrossoverPoint.setMin(0.1);
        sliderCrossoverPoint.setMax(0.9);
        sliderCrossoverPoint.setValue(crossoverRatio);
        sliderCrossoverPoint.valueProperty().addListener((observableValue, oldValue, newValue) -> {
            crossoverRatio = Math.round(newValue.doubleValue() * 100.0) / 100.0;
            labelCrossoverPoint.setText(String.format("%.0f", (crossoverRatio * 100)) + "%");
        });
    }

    /* ******************************************************************************************* */
    // Function Name: createPolygonRadioBtns                                                       //
    // Return Type: void                                                                           //
    // Parameters: None                                                                            //
    // Description: This function adds the listener for the Radio Buttons that allow the user to   //
    //              picks what type of polygons that are used in the algorithm. The choice is      //
    //              stored in the numOfVertices variable. The triangles radio button is also set   //
    //              to be selected in case the user forgets to choose one.                         //
    /* ******************************************************************************************* */
    private void createPolygonRadioBtns() {
        ToggleGroup group = new ToggleGroup();

        radioBtnTriangles.setToggleGroup(group);
        radioBtnQuads.setToggleGroup(group);

        radioBtnTriangles.setSelected(true);

        group.selectedToggleProperty().addListener((observable, oldValue, newValue) -> {
            if(newValue != null) {
                numOfVertices = (newValue.hashCode() == radioBtnTriangles.hashCode()) ? 3 : 4;
            }
        });
    }

    /* ******************************************************************************************* */
    // Function Name: createElitismButtons                                                         //
    // Return Type: void                                                                           //
    // Parameters: None                                                                            //
    // Description: This function adds the listener for the toggle buttons that allow the user to  //
    //              enable or disable elitism. The choice is stored in the elitism variable.       //
    //              Elitism is also set to on in case the user forgets to choose.                  //
    /* ******************************************************************************************* */
    private void createElitismButtons() {
        ToggleGroup group = new ToggleGroup();

        toggleOnElitism.setToggleGroup(group);
        toggleOnElitism.setSelected(elitism);

        toggleOffElitism.setToggleGroup(group);
        toggleOffElitism.setSelected(!elitism);

        group.selectedToggleProperty().addListener((observable, oldValue, newValue) -> {
            if(newValue != null) {
                elitism = newValue.hashCode() == toggleOnElitism.hashCode();
            }
        });
    }

    /* ******************************************************************************************* */
    // Function Name: createRandomInheritanceButtons                                               //
    // Return Type: void                                                                           //
    // Parameters: None                                                                            //
    // Description: This function adds the listener for the toggle buttons that allow the user to  //
    //              pick between Single-Point and Multiple-Point crossover. The choice is stored   //
    //              in the randomInheritance variable. If random inheritance is enabled, then the  //
    //              algorithm uses multiple-point crossover and the slider that lets the user      //
    //              choose the crossover ratio is disabled, else the algorithm uses single-point   //
    //              crossover and the slider that lets the user choose the crossover ratio is      //
    //              enabled.                                                                       //
    /* ******************************************************************************************* */
    private void createRandomInheritanceButtons() {
        ToggleGroup group = new ToggleGroup();
        crossoverAnchorPane.setVisible(!randomInheritance);

        toggleOnInheritance.setToggleGroup(group);
        toggleOnInheritance.setSelected(randomInheritance);

        toggleOffInheritance.setToggleGroup(group);
        toggleOffInheritance.setSelected(!randomInheritance);

        group.selectedToggleProperty().addListener((observable, oldValue, newValue) -> {
            if(newValue != null) {
                if(newValue.hashCode() == toggleOnInheritance.hashCode()) {
                    randomInheritance = true;
                    crossoverAnchorPane.setVisible(false);
                }
                else {
                    randomInheritance = false;
                    crossoverAnchorPane.setVisible(true);
                }
            }
        });
    }

    /* ******************************************************************************************* */
    // Function Name: openValidFile                                                                //
    // Return Type: boolean                                                                        //
    // Parameters: javafx.stage.Stage                                                              //
    // Description: This function is called when the user clicks on the 'Import' button. A new     //
    //              FileChooser is instantiated and opened. If the user selected an image and      //
    //              clicks 'Open' in the open file dialog, that file is saved in the selectedFile  //
    //              variable. If the user clicks cancel in the open file dialog, then selectedFile //
    //              remains null. Before the function continues, it checks if selectedFile is      //
    //              null. If it is null, the function returns false and the open file dialog       //
    //              disappears. If the user cannot select parameters until they import a valid     //
    //              image. If selectedFile is not null, the function makes sure it is some sort of //
    //              image by checking if the file name contains .jpg, .jpeg, or .png. If the image //
    //              is not one of those types, then a warning message is shown and the open file   //
    //              is reopened.                                                                   //
    /* ******************************************************************************************* */
    private boolean openValidFile(Stage stage) {
        FileChooser fileChooser = new FileChooser();
        while(true) {
            selectedFile = fileChooser.showOpenDialog(stage);
            if(selectedFile == null) {
                return false;
            }
            else {
                String fileName = selectedFile.getName();
                if (fileName.contains(".jpg") || fileName.contains(".jpeg") || fileName.contains(".png")) {
                    return true;
                }
                else {
                    Alert alert = new Alert(Alert.AlertType.WARNING);
                    alert.setContentText("File selected must be of type JPG/JPEG or PNG.");
                    alert.show();
                }
            }
        }
    }

    /* ******************************************************************************************* */
    // Function Name: convertToImage                                                               //
    // Return Type: boolean                                                                        //
    // Parameters: None                                                                            //
    // Description: This function is called if openValidFile returns true. If openValidFile        //
    //              returns true, that means that the user picked a valid image. This function     //
    //              converts selectedFile into a java.awt.image.BufferedImage which is used to     //
    //              compare generated images against pixel-by-pixel. This function also converts   //
    //              selectedFile to a javafx.scene.image.Image so that it is able to be displayed  //
    //              to the window.                                                                 //
    /* ******************************************************************************************* */
    private boolean convertToImage() {
        try {
            importedBuffIm = ImageIO.read(selectedFile);
        }
        catch(IOException ioe) {
            Alert alert = new Alert(Alert.AlertType.ERROR);
            alert.setContentText("Unable to open image.\nError: " + ioe);
            alert.show();
            return false;
        }

        double scale = (double)(importedBuffIm.getWidth()) / (double)(importedBuffIm.getHeight());
        int width = 400;
        int height = (int)(width * scale);
        imgViewImported.setFitWidth(width);
        imgViewImported.setFitHeight(height);

        Image scaledImage = SwingFXUtils.toFXImage(importedBuffIm, null);
        imgViewImported.setImage(scaledImage);

        return true;
    }
}
